import copy
import dataclasses
import json
from collections.abc import Mapping
from functools import cached_property, lru_cache
from typing import Any, Dict, List, Optional, Type, Union

import importlib_metadata
import marshmallow as ma
from marshmallow import fields

from ..utils.deepmerge import deepmerge
from ..utils.import_class import import_class
from ..utils.properties import class_property
from ..validation.utils import PermissiveSchema


@dataclasses.dataclass
class Import:
    import_path: str
    alias: Optional[str] = None

    @staticmethod
    def from_config(d):
        if isinstance(d, dict):
            return Import(d["import"], d.get("alias"))
        elif isinstance(d, (tuple, list)):
            return [Import.from_config(x) for x in d]

    def __hash__(self):
        return hash(self.import_path) ^ hash(self.alias)

    def __eq__(self, o):
        return self.import_path == o.import_path and self.alias == o.alias


@dataclasses.dataclass
class Section:
    section_name: str
    config: Dict[str, Any]
    children: Dict[str, "AbstractDataType"] = dataclasses.field(default_factory=dict)
    item: "AbstractDataType" = None

    @cached_property
    def fingerprint(self):
        f = [self.section_name, json.dumps(self.config, sort_keys=True, default=repr)]
        for key, dt in self.children.items():
            f.append(f"  @@@ {key} {type(dt).__name__}")
            f.append(
                "    "
                + getattr(dt, "section_" + self.section_name).fingerprint.replace(
                    "\n", "\n    "
                )
            )
        if self.item:
            f.append(f"### {type(self.item).__name__}")
            f.append(
                getattr(self.item, "section_" + self.section_name).fingerprint.replace(
                    "\n", "\n  "
                )
            )
        return "\n".join(f)


class MergedAttrDict(Mapping):
    def __init__(self, source, fallback):
        self._source = source
        self._fallback = fallback

    def __iter__(self):
        return iter(self.keys())

    def __getitem__(self, key):
        if key in self._source:
            return self._source[key]
        return self._fallback[key]

    def __getattr__(self, key):
        try:
            return self[key.replace("_", "-")]
        except KeyError:
            raise AttributeError(f"Attribute {key} not found")

    def __len__(self):
        return len(self.keys())

    def keys(self):
        return set(self._source.keys()).union(self._fallback.keys())


class AbstractDataType:
    parent: "AbstractDataType"
    children: Dict[str, "AbstractDataType"]

    def __init__(
        self,
        parent: "AbstractDataType",
        definition: Any,
        key: str,
        model: Any,
        schema: Any,
    ):
        self.parent = parent
        self.children = {}
        self.definition = definition
        self.key = key
        self.model = model
        self.schema = schema
        self._sections = {}
        self.skip_in_path = False

    def copy(self, without_children=False):
        ret = type(self)(
            # shallow copy of the definition to enable overwriting pieces of it
            self.parent,
            {**self.definition},
            self.key,
            self.model,
            self.schema,
        )
        ret.__dict__.update(
            {
                k: v
                for k, v in self.__dict__.items()
                if k
                not in (
                    "parent",
                    "definition",
                    "key",
                    "model",
                    "schema",
                    "_sections",
                    "children",
                    "item",
                )
            }
        )
        # clear up sections
        ret._sections = {}
        if without_children:
            if ret.children:
                ret.children = {}
            elif hasattr(ret, "item"):
                ret.item = None
        return ret

    def prepare(self, context):
        """
        Prepare the datatype. This might fill "children" property as well
        """

    def __getattr__(self, name):
        """
        datatype.json_schema maps to self.definition['json-schema'].
        If it has not been found, an empty section is returned.

        At first _process_json_schema is called on the datatype, if it exists.

        Before returning, process_json_schema method is called on all components (datatypes.components)
        with datatype and section keyword arguments
        """
        if name.startswith("section_"):
            name = name[len("section_") :]
            run_processors = True
        elif name.startswith("default_section_"):
            name = name[len("default_section_") :]
            run_processors = False
        else:
            config_key = name.replace("_", "-")
            if config_key in self.definition:
                return self.definition[config_key]

            return object.__getattribute__(self, name)

        # get the section
        section_key = name.replace("_", "-")
        if section_key in self._sections:
            return self._sections[section_key]
        config = self.definition.get(section_key, {})
        config = copy.deepcopy(config)

        # get the default from datatype
        if hasattr(self, name):
            deepmerge(config, copy.deepcopy(getattr(self, name)))

        section = Section(
            name, config, getattr(self, "children", {}), getattr(self, "item", None)
        )

        if run_processors:
            if hasattr(self, f"_process_{name}"):
                getattr(self, f"_process_{name}")(section=section)

            # call components
            datatypes.call_components(
                self,
                f"process_{name}",
                section=section,
            )
        self._sections[section_key] = section
        return section

    @cached_property
    def path(self):
        if not self.parent or not self.parent.path:
            return self.path_element or ""

        p = self.parent.path  # recursion
        if self.path_element:
            return p + f".{self.path_element}"
        return p

    @property
    def path_element(self):
        if self.skip_in_path:
            return None
        return self.key

    @cached_property
    def stack(self):
        ret = []
        p = self
        while p:
            ret.append(p)
            p = p.parent
        return tuple(reversed(ret))


class DataType(AbstractDataType):
    model_type = None

    class ModelSchema(ma.Schema):
        type = fields.String(required=True)
        required = fields.Bool()
        jsonschema = fields.Nested(PermissiveSchema)
        mapping = fields.Nested(PermissiveSchema)
        id = fields.String(
            metadata={
                "doc": "Optional id of this element (for example, for referencing the element)"
            }
        )

    def prepare(self, context):
        self.id = self.definition.get("id")
        datatypes.call_components(datatype=self, method="prepare", context=context)

    @class_property
    def validator(cls):
        validators = list(
            datatypes.call_class_components(datatype=cls, method="model_schema")
        ) + list(datatypes.get_class_components(cls, "ModelSchema"))
        validators = [x for x in validators if x]
        unique_validators = []
        for v in validators:
            if v not in unique_validators:
                unique_validators.append(v)

        class Meta:
            unknown = ma.RAISE

        ret = type(
            f"{cls.__name__}ModelValidator",
            (*unique_validators, cls.ModelSchema),
            {"Meta": Meta},
        )
        return ret

    def deep_iter(self):
        yield self

    def _process_json_schema(self, section: Section, **__kwargs):
        section.config.setdefault("type", self.definition["type"])

    def _process_mapping(self, section: Section, **__kwargs):
        section.config.setdefault("type", self.definition["type"])


class DataTypeComponent:
    eligible_datatypes = []


class DataTypes:
    @cached_property
    def datatype_map(self) -> Dict[str, DataType]:
        d = {}
        for entry in importlib_metadata.entry_points(
            group="oarepo_model_builder.datatypes"
        ):
            for dt in entry.load():
                d[dt.model_type] = dt
        return d

    @cached_property
    def components(self) -> List[DataTypeComponent]:
        c = []
        for entry in importlib_metadata.entry_points(
            group="oarepo_model_builder.datatypes.components"
        ):
            for component in entry.load():
                c.append(component())
        return c

    @lru_cache(maxsize=1024)
    def _get_components(self, datatype_class):
        datatype_components = []

        for component in self.components:
            if not component.eligible_datatypes:
                datatype_components.append(component)
            else:
                for depends_on in component.eligible_datatypes:
                    if isinstance(depends_on, str):
                        depends_on = import_class(depends_on)
                    if issubclass(datatype_class, depends_on):
                        datatype_components.append(component)
                        break

        # remove overridden components
        unsorted_components = []
        non_leaf_components = set()
        for c in datatype_components:
            non_leaf_components.update(type(c).mro()[1:])

        for c in datatype_components:
            if type(c) not in non_leaf_components:
                unsorted_components.append(c)

        dependency_remaps = {}
        for c in unsorted_components:
            remap = getattr(c, "dependency_remap", None)
            if remap:
                dependency_remaps[remap] = type(c)
        # sort by dependencies
        depsort_map = {}

        def get_dependent_component(component):
            return (
                dependency_remaps[component]
                if component in dependency_remaps
                else component
            )

        for c in unsorted_components:
            dependencies_classes = depsort_map.setdefault(type(c), [])
            depsort_map[type(c)] = dependencies_classes
            for depends_on in getattr(c, "depends_on", []):
                depends_on = get_dependent_component(depends_on)
                if isinstance(depends_on, str):
                    depends_on = import_class(depends_on)
                dependencies_classes.append(depends_on)
            for affects in getattr(c, "affects", []):
                affects = get_dependent_component(affects)
                if isinstance(affects, str):
                    affects = import_class(affects)
                depsort_map.setdefault(affects, []).append(type(c))

        sorted_components = []
        while unsorted_components:
            new_unsorted_components = []
            current_round_sorted_components = set()
            for c in unsorted_components:
                if not depsort_map[type(c)]:
                    sorted_components.append(c)
                    current_round_sorted_components.add(type(c))
                else:
                    new_unsorted_components.append(c)
            if len(new_unsorted_components) == len(unsorted_components):
                raise AttributeError(
                    f"A loop has been detected in component dependencies: {unsorted_components}"
                )
            for c in new_unsorted_components:
                depsort_map[type(c)] = [
                    x
                    for x in depsort_map[type(c)]
                    if x not in current_round_sorted_components
                ]
            unsorted_components = new_unsorted_components
        return tuple(sorted_components)

    def get_datatype(self, parent, data, key, model, schema) -> Union[DataType, None]:
        datatype_class = self.get_datatype_class(data.get("type", None))
        if datatype_class:
            return datatype_class(parent, data, key, model, schema)
        if parent:
            raise KeyError(
                f"Do not have datatype for the following data at path '{parent.path}':\n"
                f"{json.dumps(data, indent=4, ensure_ascii=False)}"
            )
        raise KeyError(
            f"Do not have datatype for the following data:\n"
            f"{json.dumps(data, indent=4, ensure_ascii=False)}"
        )

    def get_datatype_class(self, datatype_type):
        return self.datatype_map.get(datatype_type)

    def call_components(self, datatype: DataType, method: str, **kwargs):
        ret = []
        for component in self._get_components(type(datatype)):
            if hasattr(component, method):
                ret.append(getattr(component, method)(datatype=datatype, **kwargs))
        return ret

    def call_class_components(self, datatype: Type[DataType], method: str, **kwargs):
        ret = []
        for component in self._get_components(datatype):
            if hasattr(component, method):
                ret.append(getattr(component, method)(datatype=datatype, **kwargs))
        return ret

    def get_class_components(self, datatype: Type[DataType], name: str):
        ret = []
        for component in self._get_components(datatype):
            if hasattr(component, name):
                ret.append(getattr(component, name))
        return ret

    def _clear_caches(self):
        """
        mostly for testing
        """
        self._get_components.cache_clear()
        try:
            del self.datatype_map
        except:  # NOSONAR intentionally broad, used only in tests
            pass
        try:
            del self.components
        except:  # NOSONAR intentionally broad, used only in tests
            pass


datatypes = DataTypes()



import os

import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType

from .defaults import DefaultsModelComponent
from .utils import set_default


class SavedModelSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    file_ = ma.fields.Str(
        attribute="file",
        data_key="file",
        metadata={"doc": "File where to save the model"},
    )
    alias = ma.fields.String(
        metadata={"doc": "Alias under which the model is registered in setup.cfg"}
    )
    module = ma.fields.String(metadata={"doc": "Module where the file is saved"})


class SavedModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        saved_model = ma.fields.Nested(
            SavedModelSchema, attribute="saved-model", data_key="saved-model"
        )

    def before_model_prepare(self, datatype, context=None, **kwargs):
        profile = context.get("profile_module")
        file = f"{profile}.json"

        module_path = datatype.definition["module"]["path"]
        module = datatype.definition["module"]["qualified"]

        saved_model = set_default(datatype, "saved-model", {})

        saved_model.setdefault(
            "file",
            os.path.join(module_path, "models", file),
        )
        saved_model.setdefault(
            "module",
            f"{module}.models",
        )
        saved_model.setdefault("alias", datatype.definition["module"]["alias"])



from typing import List

from ...datatypes import DataType, DataTypeComponent
from .field import RegularMarshmallowComponentMixin
from .graph import MarshmallowField


class UIMarshmallowComponent(RegularMarshmallowComponentMixin, DataTypeComponent):
    def ui_marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **kwargs
    ):
        marshmallow = datatype.section_ui
        self._create_marshmallow_field(
            datatype,
            marshmallow,
            marshmallow.config.setdefault("marshmallow", {}),
            fields,
            **kwargs,
        )



import keyword
import os
import re


def convert_name_to_python(name):
    if not name:
        return name

    # special case :)
    name = name.replace("@v", "_version")
    if name == "id":
        name = name.replace("id", "_id")
    # Replace any spaces or special characters in the string with an underscore
    identifier = re.sub(r"[^\w\s]", "_", name)
    identifier = re.sub(r"\s+", "_", identifier)

    # Ensure that the first character of the identifier is a letter or an underscore
    if identifier[0].isdigit():
        identifier = "_" + identifier

    # Ensure that the resulting identifier is not a reserved keyword in Python
    if keyword.iskeyword(identifier):
        identifier = identifier + "_"

    return identifier


def capitalize(s):
    return s[0].capitalize() + s[1:]


def convert_name_to_python_class(name):
    class_name = "".join([capitalize(word) for word in name.split("-")])
    class_name = "".join([capitalize(word) for word in class_name.split("_")])

    # Replace any spaces or special characters in the string with an underscore
    class_name = re.sub(r"[^\w\s]", "", class_name)
    class_name = re.sub(r"\s+", "_", class_name)

    if class_name[0].isdigit():
        class_name = "_" + class_name

    return class_name


def package_name(value):
    if not value:
        return None
    return value.rsplit(".", maxsplit=1)[0]


def split_package_name(value):
    return package_name(value)


def base_name(value):
    if not value:
        return None
    return value.rsplit(".", maxsplit=1)[-1]


def split_base_name(value):
    return base_name(value)


def split_package_base_name(value):
    if "." not in value:
        return None, value
    return value.rsplit(".", maxsplit=1)


def qualified_name(package_name: str, class_name: str):
    if "." not in class_name:
        return f"{package_name}.{class_name}"
    if class_name.startswith("."):
        # package_name: aaa.bbb, class name .C => aaa.bbb.C
        # package_name: aaa.bbb, class name ..C => aaa.C
        # package_name: aaa.bbb, class name ...C => C
        class_name = class_name[1:]
        package_path = package_name.split(".")
        while class_name.startswith("."):
            if package_path:
                package_path = package_path[:-1]
            class_name = class_name[1:]
        if package_path:
            class_name = f"{'.'.join(package_path)}.{class_name}"
    return class_name


def convert_config_to_qualified_name(
    config_section, module_field="module", name_field="class"
):
    module = config_section[module_field]
    class_name = config_section[name_field]
    qualified = qualified_name(module, class_name)
    if qualified != class_name:
        config_section[name_field] = qualified


def module_to_path(module):
    return os.path.join(*module.split("."))


def parent_module(module):
    return ".".join(module.split(".")[:-1])



import logging

from libcst import Assign, Integer

from .common import IdentityBaseMerger, IdentityMerger, MergerBase, PythonContext
from .mergers import expression_mergers, simple_line_mergers

log = logging.getLogger("oarepo_model_builder.cst")


class AssignMerger(MergerBase):
    def merge_internal(self, context: PythonContext, existing_node, new_node):
        merger = expression_mergers.get(
            type(existing_node.value or new_node.value), IdentityMerger()
        )
        merged_value = merger.merge(context, existing_node.value, new_node.value)
        return existing_node.with_changes(value=merged_value)

    def identity(self, context, node):
        return Assign(targets=node.targets, value=Integer(value="1"))


class ImportMerger(IdentityBaseMerger):
    pass


class ImportFromMerger(IdentityBaseMerger):
    pass


class ExprMerger(IdentityBaseMerger):
    pass


class PassMerger(IdentityBaseMerger):
    pass


class IntegerMerger(IdentityBaseMerger):
    pass


class SimpleStringMerger(IdentityBaseMerger):
    pass


class NameMerger(IdentityBaseMerger):
    pass


class StarredElementMerger(IdentityBaseMerger):
    pass


class FunctionMerger(MergerBase):
    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return existing_node

    def identity(self, context, node):
        return node.name


class SimpleStatementLineMerger(MergerBase):
    def identity(self, context, node):
        assert len(node.body) == 1
        body = node.body[0]
        if type(body) in simple_line_mergers:
            return simple_line_mergers[type(body)].identity(context, body)
        log.error("Could not find node %s in simple_line_mergers", type(node))
        return node.body[0]

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        existing_body = existing_node.body[0] if existing_node else None
        new_body = new_node.body[0] if new_node else None
        merger = simple_line_mergers.get(
            type(existing_body or new_body), IdentityMerger()
        )
        return existing_node.with_changes(
            body=[merger.merge(context, existing_body, new_body)]
        )



import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.utils.facet_helpers import facet_name, flatten
from oarepo_model_builder.validation.utils import ImportSchema

from ...datatypes import DataTypeComponent
from . import FacetDefinition


class FacetsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    searchable = ma.fields.Bool(required=False)
    key = fields.String(required=False)
    field = fields.String(required=False)
    facet_class = fields.String(
        required=False, data_key="facet-class", attribute="facet-class"
    )
    args = fields.List(fields.String(), required=False)
    imports = fields.List(fields.Nested(ImportSchema), required=False)
    path = fields.String(required=False)
    keyword = fields.String(required=False)
    facet_groups = fields.List(fields.String(), required=False, data_key="facet-groups", attribute="facet-groups")
    facet = ma.fields.Bool(required=False)

class RegularFacetsComponent(DataTypeComponent):
    eligible_datatypes = []

    class ModelSchema(ma.Schema):
        facets = ma.fields.Nested(
            FacetsSchema,
            required=False,
        )

    def facet_path(self, datatype, facet_section):
        _path = datatype.path
        if "path" in facet_section:
            path = _path + "." + facet_section["path"]
        else:
            path = _path
        return path

    def process_facets(self, datatype, section, **__kwargs):
        # create the facet definition
        facet_section = section.config

        path = self.facet_path(datatype, facet_section)
        facet_definition = FacetDefinition(
            path=facet_section.get("key", facet_name(datatype.path)),
            dot_path=datatype.path,
            searchable=facet_section.get("searchable"),
            imports=facet_section.get("imports", []),
            facet=facet_section.get("facet", None),
            facet_groups= facet_section.get("facet-groups", ["default"])
        )

        # set the field on the definition
        label = facet_section.get("label", f'{datatype.path.replace(".", "/")}.label')
        facet_definition.set_field(
            facet_section,
            arguments=[
                f"field={repr(path)}",
                f"label =_({repr(label)})",
                *facet_section.get("args", []),
            ],
        )

        # if there is indeed a facet here, process via parents
        if facet_definition.field:
            facets = flatten(
                datatypes.call_components(
                    datatype.parent,
                    "build_facet_definition",
                    facet_definition=facet_definition,
                )
            )
        else:
            facets = []

        # and store it on the element
        section.config["facets"] = facets
        return section



from .app import AppModelComponent
from .blueprints import BlueprintsModelComponent
from .defaults import DefaultsModelComponent
from .ext_resource import ExtResourceModelComponent
from .facets import FacetsModelComponent
from .jsonschema import JSONSchemaModelComponent
from .mapping import MappingModelComponent
from .marshmallow import MarshmallowModelComponent
from .model_saver import SavedModelComponent
from .permissions import PermissionsModelComponent
from .pid import PIDModelComponent
from .proxy import ProxyModelComponent
from .record import RecordModelComponent
from .record_dumper import RecordDumperModelComponent
from .record_metadata import RecordMetadataModelComponent
from .resource import ResourceModelComponent
from .sample import SampleModelComponent
from .search_options import SearchOptionsModelComponent
from .service import ServiceModelComponent
from .sort import SortModelComponent
from .ui import UIModelComponent
from .ui_marshmallow import UIMarshmallowModelComponent

__all__ = [
    "AppModelComponent",
    "BlueprintsModelComponent",
    "FacetsModelComponent",
    "SearchOptionsModelComponent",
    "JSONSchemaModelComponent",
    "MappingModelComponent",
    "MarshmallowModelComponent",
    "PIDModelComponent",
    "PermissionsModelComponent",
    "RecordDumperModelComponent",
    "RecordMetadataModelComponent",
    "RecordModelComponent",
    "ResourceModelComponent",
    "SampleModelComponent",
    "SavedModelComponent",
    "ServiceModelComponent",
    "UIMarshmallowModelComponent",
    "UIModelComponent",
    "DefaultsModelComponent",
    "ProxyModelComponent",
    "ExtResourceModelComponent",
    "SortModelComponent",
]



import marshmallow as ma

from oarepo_model_builder.datatypes import ModelDataType
from oarepo_model_builder.datatypes.components.model.utils import set_default

from ..facets import FacetDefinition
from ..facets.object import ObjectFacetsComponent
from .defaults import DefaultsModelComponent


class FacetsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    searchable = ma.fields.Bool(required=False)

    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be pasted to search options module"},
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the facets will be placed"}
    )
    generate = ma.fields.Boolean()
    skip = ma.fields.Boolean()


class FacetsModelComponent(ObjectFacetsComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        facets = ma.fields.Nested(
            FacetsSchema,
            required=False,
        )

    def before_model_prepare(self, datatype, *, context, **__kwargs):
        module = datatype.definition["module"]["qualified"]
        profile_module = context["profile_module"]

        facets = set_default(datatype, "facets", {})
        facets.setdefault("generate", True)
        facets.setdefault("module", f"{module}.services.{profile_module}.facets")

        facets.setdefault("extra-code", "")

    def build_facet_definition(
        self,
        datatype,
        facet_definition: FacetDefinition,
    ):
        if facet_definition.searchable is None:
            facet_definition.searchable = datatype.definition.get("searchable", True)
        if facet_definition.searchable is not False and facet_definition.facet is not False:
            return [facet_definition]
        else:
            # facet will not be generated
            return []



from oarepo_model_builder.datatypes import ObjectDataType, datatypes
from oarepo_model_builder.utils.facet_helpers import flatten

from . import FacetDefinition
from .field import RegularFacetsComponent


class ObjectFacetsComponent(RegularFacetsComponent):
    eligible_datatypes = [ObjectDataType]

    def build_facet_definition(
        self,
        datatype,
        facet_definition: FacetDefinition,
    ):
        facet_definition.update(datatype.section_facets.config)

        return flatten(
            datatypes.call_components(
                datatype.parent,
                "build_facet_definition",
                facet_definition=facet_definition,
            )
        )



import marshmallow as ma

from .extensibility import ExtensibleSchema
from .plugins import PluginsSchema


def get_model_schema():
    from ..datatypes.model import ModelDataType

    return ModelDataType.validator()


class DefsSchema(ma.Schema):
    class Meta:
        unknown = ma.INCLUDE


class SettingsPythonSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    use_isort = ma.fields.Bool(
        attribute="use-isort", data_key="use-isort", load_default=True
    )
    use_black = ma.fields.Bool(
        attribute="use-black", data_key="use-black", load_default=True
    )
    use_autoflake = ma.fields.Bool(
        attribute="use-autoflake", data_key="use-autoflake", load_default=True
    )


class SettingsOpenSearchSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    version = ma.fields.String(load_default="os-v2")


class SettingsSchema(ma.Schema):
    python = ma.fields.Nested(ExtensibleSchema("settings.python", SettingsPythonSchema))
    opensearch = ma.fields.Nested(
        ExtensibleSchema("settings.opensearch", SettingsOpenSearchSchema)
    )
    schema_server = ma.fields.String(
        attribute="schema-server", data_key="schema-server", load_default="local://"
    )
    oarepo_version = ma.fields.String(
        attribute="oarepo-version", data_key="oarepo-version", load_default="11"
    )
    extension_elements = ma.fields.List(
        ma.fields.String(),
        attribute="extension-elements",
        data_key="extension-elements",
    )

    class Meta:
        unknown = ma.RAISE


class ModelFileSchema(ma.Schema):
    version = ma.fields.Str(
        load_default="1.0.0", metadata={"doc": "Model version, default value 1.0.0"}
    )
    record = ma.fields.Nested(get_model_schema, metadata={"doc": "Main record"})
    defs = ma.fields.Nested(
        DefsSchema,
        attribute="$defs",
        data_key="$defs",
        metadata={"doc": "Extra definitions, might be included via _use_ or _extend_"},
    )
    settings = ma.fields.Nested(
        ExtensibleSchema("settings", SettingsSchema),
        metadata={"doc": "General settings, applies to all generated sources"},
        load_default=lambda: SettingsSchema().load({}),
    )
    runtime_dependencies = ma.fields.Dict(
        ma.fields.String(),
        ma.fields.String(),
        attribute="runtime-dependencies",
        data_key="runtime-dependencies",
    )
    dev_dependencies = ma.fields.Dict(
        ma.fields.String(),
        ma.fields.String(),
        attribute="dev-dependencies",
        data_key="dev-dependencies",
    )
    plugins = ma.fields.Nested(
        ExtensibleSchema("plugins", PluginsSchema),
        metadata={"doc": "Plugins to load, enable, disable"},
    )
    profiles = ma.fields.List(ma.fields.String())


class ModelValidator:
    def validate(self, data):
        complete_model_schema = ExtensibleSchema("model_file", ModelFileSchema)
        validator = complete_model_schema()
        return validator.load(data)


model_validator = ModelValidator()



class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def class_property(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


class CachedClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        cache_name = f"_cache_class_property_{klass.__name__}"
        if hasattr(klass, cache_name):
            return getattr(klass, cache_name)
        ret = self.fget.__get__(obj, klass)()
        setattr(klass, cache_name, ret)
        return ret

    def __delete__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        cache_name = f"_cache_class_property_{klass.__name__}"
        if hasattr(klass, cache_name):
            return delattr(klass, cache_name)

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")


def cached_class_property(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return CachedClassPropertyDescriptor(func)



import marshmallow as ma
from marshmallow import fields
from marshmallow.exceptions import ValidationError


class CheckedConstant(fields.Constant):
    def _deserialize(self, value, *args, **kwargs):
        if value != self.constant:
            raise ValidationError(f"Bad value: '{self.constant}' expected.")
        return super()._deserialize(value, *args, **kwargs)


class ExtendablePartSchema(ma.Schema):
    extend = fields.String(required=False)
    use = fields.String(required=False)


class PermissiveSchema(ExtendablePartSchema):
    class Meta:
        unknown = ma.INCLUDE

    def dump(self, obj, *, many=None):
        if not obj:
            return super().dump(obj, many=many)
        if many:
            return [self.dump(x, many=False) for x in obj]
        return {**obj, **super().dump(obj, many=False)}


class StrictSchema(ExtendablePartSchema):
    class Meta:
        unknown = ma.RAISE


class ImportSchema(ExtendablePartSchema):
    _import = fields.String(data_key="import", attribute="import", required=True)
    alias = fields.String(required=False)



import dataclasses
from typing import List, Optional, Set

from oarepo_model_builder.validation import InvalidModelException

from ....utils.python_name import base_name, package_name
from ...datatypes import Import


@dataclasses.dataclass
class MarshmallowField:
    key: str
    definition: str
    imports: List[Import] = dataclasses.field(default_factory=list)
    owner: "MarshmallowClass" = None
    reference: "MarshmallowReference" = None

    @property
    def full_definition(self):
        if not self.reference:
            return self.definition
        return self.reference.apply(self.definition)


@dataclasses.dataclass
class MarshmallowReference:
    reference: str
    referenced_class: "MarshmallowClass" = None
    accessor: Optional[str] = None

    def set_class(self, classes):
        if self.reference in classes:
            self.referenced_class = classes[self.reference]

    def apply(self, definition):
        if self.accessor:
            ref = self.accessor
        else:
            ref = f"lambda: {base_name(self.reference)}()"

        return definition.replace("__reference__", ref)


@dataclasses.dataclass
class MarshmallowClass:
    class_name: str
    base_classes: List[str]
    imports: List[Import]
    fields: List[MarshmallowField]
    unknown: str
    order: Optional[int] = None

    references: List[MarshmallowReference] = dataclasses.field(default_factory=list)

    def collect_references(self, classes):
        references = []
        for fld in self.fields:
            if fld.reference and fld.reference.reference in classes:
                fld.reference.set_class(classes)
                references.append(fld.reference)
        self.references = references

    def remove_references(self, class_names):
        self.references = list(
            filter(lambda r: r.reference not in class_names, self.references)
        )


def sort_by_reference_count(classes_list: List[MarshmallowClass]):
    classes = {x.class_name: x for x in classes_list}
    to_process = list(classes_list)
    processed_class_names = set()

    for c in classes_list:
        c.collect_references(classes)

    order = 0
    while to_process:
        order += 1
        for_another_round = []
        for c in to_process:
            if c.references:
                for_another_round.append(c)
            else:
                processed_class_names.add(c.class_name)
                c.order = order
        if len(for_another_round) == len(to_process):
            raise InvalidModelException("Cycle found, code missing")
        else:
            for c in for_another_round:
                c.remove_references(processed_class_names)
        to_process = for_another_round
    classes_list.sort(key=lambda x: [-x.order, x.class_name])


def collect_imports(current_module, classes_list: List[MarshmallowClass]):
    for cls in classes_list:
        for fld in cls.fields:
            if not fld.reference or not fld.reference.reference:
                continue
            if fld.reference.accessor:
                # local accessor handles its own imports inside the accessor
                continue
            # if from different package,
            if package_name(fld.reference.reference) != current_module:
                fld.imports.append(Import(import_path=fld.reference.reference))


@dataclasses.dataclass
class PackageWithDependencies:
    package: str
    direct_dependencies: Set[str] = dataclasses.field(default_factory=set)
    accessor_dependencies: Set[str] = dataclasses.field(default_factory=set)
    dependencies: Set[str] = dataclasses.field(default_factory=set)

    def add_dependency(self, dependency: str):
        self.dependencies.add(dependency)

    def remove_dependencies(self, dependencies):
        for dep in dependencies:
            if dep in self.dependencies:
                self.direct_dependencies.add(dep)
                self.dependencies.remove(dep)

    def break_dependencies(self):
        self.accessor_dependencies.update(self.dependencies)
        self.dependencies.clear()


def set_package_dependencies(classes_by_package):
    package_dependencies: List[PackageWithDependencies] = []
    for p in classes_by_package:
        pd = PackageWithDependencies(p)
        package_dependencies.append(pd)
        for pc in classes_by_package[p]:
            for fld in pc.fields:
                if not fld.reference:
                    continue
                referenced_package = package_name(fld.reference.reference)
                if referenced_package != p and referenced_package in classes_by_package:
                    pd.add_dependency(referenced_package)
    to_process = list(package_dependencies)
    processed_dependencies = set()
    while to_process:
        for_another_round = []
        for c in to_process:
            if c.dependencies:
                for_another_round.append(c)
            else:
                processed_dependencies.add(c.package)
        if len(for_another_round) == len(to_process):
            raise InvalidModelException(
                f"Cycle found in packages, code missing: {to_process}"
            )
        else:
            for c in for_another_round:
                c.remove_dependencies(processed_dependencies)
        to_process = for_another_round

    for p in package_dependencies:
        for cls in classes_by_package[p.package]:
            for fld in cls.fields:
                if not fld.reference:
                    continue
                reference_package_name = package_name(fld.reference.reference)
                if reference_package_name in p.direct_dependencies:
                    fld.imports.append(Import(fld.reference.reference))
                elif reference_package_name in p.accessor_dependencies:
                    fld.reference.accessor = f"get{base_name(fld.reference.reference)}"



from typing import List

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder
from .invenio_record_facets import get_distinct_facets


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search_options"
    section = "search-options"
    template = "record-search-options"

    def finish(self, **extra_kwargs):
        facets: List[FacetDefinition] = get_distinct_facets(self.current_model)
        facet_groups = {}
        default_group = []
        search_data = []
        for f in facets:
            for group in f.facet_groups:
                if group != 'default':
                    if group not in facet_groups.keys():
                        facet_groups[group] = {}
                    facet_groups[group][f.path] = "facets." + f.path
                if group == 'default':
                    default_group.append({f.path: "facets." + f.path})
            search_data.append({f.path: "facets." + f.path})
        if "sortable" in self.current_model.definition:
            sort_options = self.current_model.definition["sortable"]
        else:
            sort_options = {}
        extra_kwargs["facet_groups"] = facet_groups
        extra_kwargs["default_group"] = default_group
        extra_kwargs["sort_definition"] = sort_options
        super().finish(**extra_kwargs)



import copy
from typing import List

import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import DataTypeComponent
from oarepo_model_builder.utils.python_name import convert_name_to_python
from oarepo_model_builder.validation.utils import ImportSchema, StrictSchema

from ...datatypes import DataType, Import
from .graph import MarshmallowField


class PropertyMarshmallowSchema(StrictSchema):
    read = fields.Boolean(required=False)
    write = fields.Boolean(required=False)
    imports = fields.List(fields.Nested(ImportSchema), required=False)
    field_name = fields.String(
        data_key="field-name", attribute="field-name", required=False
    )
    field = fields.String(required=False)
    field_class = fields.String(
        data_key="field-class", attribute="field-class", required=False
    )
    arguments = fields.List(fields.String(), required=False)
    validators = fields.List(fields.String(), required=False)


class RegularMarshmallowComponentMixin:
    def _create_marshmallow_field(
        self, datatype, section, marshmallow, fields, **kwargs
    ):
        read = marshmallow.get("read", True)
        write = marshmallow.get("write", True)
        if not read and not write:
            return

        imports = Import.from_config(marshmallow.get("imports", []))
        field = marshmallow.get("field", None)

        key = datatype.key
        if key:
            field_name = marshmallow.get("field-name")
            if not field_name:
                field_name = convert_name_to_python(datatype.key)
                marshmallow["field-name"] = field_name
        else:
            field_name = None

        if not field:
            field_class = marshmallow.get("field-class")
            if not field_class:
                return
            field_decl = [
                field_class,
                "(",
                ", ".join(
                    self._marshmallow_field_arguments(
                        datatype, section, marshmallow, field_name, **kwargs
                    )
                ),
                ")",
            ]
            field = "".join(field_decl)

        fields.append(MarshmallowField(field_name, field, imports))

    def _marshmallow_field_arguments(
        self, datatype, section, marshmallow, field_name  # NOSONAR
    ):
        arguments = copy.copy(marshmallow.get("arguments", []))
        required = datatype.definition.get("required", False)
        if required:
            arguments.append("required=True")
        read = marshmallow.get("read", True)
        write = marshmallow.get("write", True)
        if read and not write:
            arguments.append("dump_only=True")
        elif write and not read:
            arguments.append("load_only=True")

        key = datatype.key

        if key != field_name:
            arguments.append(f'data_key="{key}"')
            arguments.append(f'attribute="{key}"')

        validators = marshmallow.get("validators", [])
        if validators:
            arguments.append(f"validate=[{', '.join(validators)}]")

        return arguments


class RegularMarshmallowComponent(RegularMarshmallowComponentMixin, DataTypeComponent):
    class ModelSchema(ma.Schema):
        marshmallow = ma.fields.Nested(
            PropertyMarshmallowSchema,
            required=False,
        )

    def marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **kwargs
    ):
        marshmallow = datatype.section_marshmallow
        self._create_marshmallow_field(
            datatype, marshmallow, marshmallow.config, fields, **kwargs
        )



from difflib import Differ

from oarepo_model_builder.outputs import OutputBase


class DiffOutput(OutputBase):
    TYPE = "diff"

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.text = self.original_data = f.read()
        except FileNotFoundError:
            self.original_data = None
            self.text = ""

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        if self.text != self.original_data:
            if self.original_data:
                lines = [
                    x[2:]
                    for x in Differ().compare(
                        self.original_data.splitlines(), self.text.splitlines()
                    )
                    if x[0] != "?"
                ]
            else:
                lines = [self.text]

            with self.builder.filesystem.open(self.path, "w") as f:
                f.write("\n".join(lines))

    def write(self, text):
        self.text = text



import logging

import marshmallow as ma
import yaml
from marshmallow.exceptions import ValidationError
from marshmallow_oneofschema import OneOfSchema

from ..datatypes import DataType

log = logging.getLogger("datatypes")


class PropertySchemas:
    def __init__(self, is_array=False) -> None:
        self.is_array = is_array

    def get(self, property_type):
        from ..datatypes import datatypes

        datatype = datatypes.get_datatype_class(property_type)
        if not datatype:
            raise ma.exceptions.ValidationError(
                f"No datatype registered for {property_type}"
            )
        return datatype.validator


class FieldSchema(OneOfSchema):
    type_field_remove = False
    type_schemas = PropertySchemas(is_array=False)

    def get_obj_type(self, obj):
        return obj["type"]

    def load(self, data, *, many=None, partial=None, unknown=None, **kwargs):
        # OneOfSchema does not call pre-load actions, so add it here explicitly
        if self._has_processors(ma.decorators.PRE_LOAD):
            try:
                processed_data = self._invoke_load_processors(
                    ma.decorators.PRE_LOAD,
                    data,
                    many=many,
                    original_data=data,
                    partial=partial,
                )
            except ValidationError as err:
                errors = err.normalized_messages()
                exc = ValidationError(errors, data=data, valid_data={})
                self.handle_error(exc, data, many=many, partial=partial)
                raise exc
        else:
            processed_data = data

        return super().load(
            processed_data, many=many, partial=partial, unknown=unknown, **kwargs
        )

    @ma.pre_load(pass_many=False)
    def set_type(self, data, **kwargs):
        if not isinstance(data, dict):
            raise ValidationError(f'Must be an object, is "{repr(data)[:20]}..."')
        if "type" not in data:
            if "properties" in data:
                data["type"] = "object"
            if "items" in data:
                data["type"] = "array"
        return data


class ObjectPropertiesField(ma.fields.Dict):
    def __init__(self, **kwargs):
        super().__init__(
            keys=ma.fields.Str(), values=ma.fields.Nested(FieldSchema), **kwargs
        )

    def deserialize(self, value, attr=None, data=None, **kwargs):
        if value is ma.missing:
            value = {}
        if not isinstance(value, dict):
            raise ma.exceptions.ValidationError(
                "Object properties field must be a mapping"
            )
        returned_value = {}
        for k, v in value.items():
            if isinstance(v, str):
                # single-line type followed by yaml => read it
                if "{" not in v:
                    v = {"type": v}
                else:
                    datatype, constraints = v.split("{", maxsplit=1)
                    constraints = constraints.replace(":", ": ")
                    v = {
                        "type": datatype,
                        **yaml.safe_load("blah: {" + constraints)["blah"],
                    }
            if k.endswith("{}"):
                k = k[:-2]
                children, top_props = self.elevate_props(v)
                v = {"type": "object", **top_props, "properties": children}
            if k.endswith("{nested}"):
                k = k[:-8]
                children, top_props = self.elevate_props(v)
                v = {"type": "nested", **top_props, "properties": children}
            if k.endswith("[]"):
                k = k[:-2]
                children, top_props = self.elevate_props(v)
                v = {"type": "array", **top_props, "items": children}
            returned_value[k] = v
        try:
            return super().deserialize(returned_value, attr=attr, data=data, **kwargs)
        except ValidationError as e:
            components = e.messages["components"]
            new_components = components.pop("value", {})
            components.update(new_components)
            raise e

    def elevate_props(self, v):
        children = {}
        top_props = {}
        for k, v in v.items():
            if k[0] == "^":
                top_props[k[1:]] = v
            else:
                children[k] = v
        return children, top_props


class ObjectDataType(DataType):
    model_type = "object"

    ui = {
        "marshmallow": {
            "field-class": "ma.fields.Nested",
        }
    }
    marshmallow = {
        "field-class": "ma.fields.Nested",
    }
    json_schema = {"type": "object"}
    mapping = {"type": "object"}

    class ModelSchema(DataType.ModelSchema):
        properties = ObjectPropertiesField()

        @ma.decorators.pre_load()
        def before_load(self, value, **kwargs):
            if not value:
                return value
            if isinstance(value, dict) and value.get("properties") is None:
                value["properties"] = {}
            return value

    def prepare(self, context):
        from ..datatypes import datatypes

        # get children
        properties = self.definition.get("properties", {})
        # children are filled in alphabet order to make them stable
        self.children = {
            k: datatypes.get_datatype(self, v, k, self.model, self.schema)
            for k, v in sorted(properties.items())
        }
        for v in self.children.values():
            v.prepare(context)

        super().prepare(context)

    def deep_iter(self):
        yield from super().deep_iter()
        for c in self.children.values():
            yield from c.deep_iter()



from oarepo_model_builder.datatypes import Import

from .invenio_record_marshmallow import InvenioRecordMarshmallowBuilder


class InvenioRecordUIMarshmallowBuilder(InvenioRecordMarshmallowBuilder):
    TYPE = "invenio_record_ui_schema"
    template = "marshmallow"
    extra_imports = [Import("oarepo_runtime.ui.marshmallow.InvenioUISchema")]

    build_class_method = "ui_marshmallow_build_class"



from jinja2 import pass_context

from ..datatypes import Import
from .python_name import (  # noqa
    base_name,
    package_name,
    split_base_name,
    split_package_base_name,
    split_package_name,
)


def sorted_imports(imports):
    # make the entries unique
    imports_dict = {(x["import"], x.get("alias")): x for x in imports}
    imports = list(imports_dict.values())
    # sort them

    imports.sort(key=lambda x: (x["import"], x.get("alias") or ""))
    return imports


def to_import_dict(import_object):
    if isinstance(import_object, Import):
        return {"import": import_object.import_path, "alias": import_object.alias}
    return import_object


@pass_context
def generate_import(ctx, import_object, imported_part="class", alias=None, skip=False):
    if skip:
        return ""
    import_object = to_import_dict(import_object)
    if isinstance(import_object, list):
        import_object = [to_import_dict(x) for x in import_object]
        return "\n".join(
            [generate_import(ctx, x, None) for x in sorted_imports(import_object)]
        )
    if isinstance(import_object, str):
        return generate_import(ctx, {"import": import_object, "alias": alias}, None)
    if imported_part:
        if import_object.get("skip"):
            return ""
        return generate_import(ctx, import_object[imported_part], alias)
    import_name = import_object["import"]
    alias = import_object.get("alias")
    ret = []

    if import_name == alias:
        # if import is the same as alias, keep just the import statement
        ret.append(f"import {import_name}")
    else:
        # otherwise break the import to path and base name and alias if defined
        if "." in import_name:
            import_path, import_name = import_name.rsplit(".", maxsplit=1)
        else:
            import_path = None
        if import_path:
            if ctx["current_module"] == import_path:
                return ""  # do not import from the same module
            ret.append(f"from {import_path} import {import_name}")
        else:
            ret.append(f"import {import_name}")
        if alias:
            ret.append(f"as {alias}")
    return " ".join(ret)


def generate_list(data, separator=", ", start=False, end=False):
    if not data:
        return ""
    ret = []
    if start:
        ret.append(separator)
    for di, d in enumerate(data):
        if di:
            ret.append(separator)
        ret.append(d)
    if end:
        ret.append(separator)
    return "".join(ret)


def in_different_package(current_module, value):
    return current_module != package_name(value)


def with_defined_prefix(always_defined_import_prefixes, value):
    return package_name(value) in always_defined_import_prefixes


def class_header(rec, class_name="class", base_classes_name="base-classes"):
    try:
        ret = [
            base_name(rec[class_name]),
        ]
    except KeyError:
        raise KeyError(f'Do not have "{class_name}" inside {rec}')
    base_classes = rec[base_classes_name]
    if base_classes:
        ret.append("(")
        for idx, cls in enumerate(base_classes):
            if idx > 0:
                ret.append(", ")
            ret.append(cls)
        ret.append(")")
    return "".join(ret)



from typing import List

from ... import ArrayDataType
from ...datatypes import DataType
from .array import ArrayMarshmallowComponentMixin
from .graph import MarshmallowField
from .ui_field import UIMarshmallowComponent


class UIArrayMarshmallowComponent(
    ArrayMarshmallowComponentMixin, UIMarshmallowComponent
):
    eligible_datatypes = [ArrayDataType]

    def ui_marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **kwargs
    ):
        section = datatype.section_ui
        self._create_marshmallow_field(
            datatype,
            section,
            section.config.setdefault("marshmallow", {}),
            fields,
            field_accessor="ui_marshmallow_field",
            **kwargs,
        )



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import convert_config_to_qualified_name
from oarepo_model_builder.validation.utils import ImportSchema

from .app import AppModelComponent
from .defaults import DefaultsModelComponent
from .utils import append_array, set_default


class ServiceClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Generate service class (default)"})
    config_key = ma.fields.Str(
        metadata={"doc": "Key under which actual service class is registered in config"}
    )
    proxy = ma.fields.Str(
        metadata={"doc": "name of the service proxy, will be put to _proxies_ package"}
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the service class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put below the generated service class"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class ServiceConfigClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    service_id = ma.fields.String(
        attribute="service-id",
        data_key="service-id",
        metadata={"doc": "ID of the service"},
    )
    generate = ma.fields.Bool(metadata={"doc": "Generate the service config"})
    config_key = ma.fields.Str(
        metadata={
            "doc": "Key under which the actual service config is registered in config"
        }
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the service config class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put below the service config class"},
    )
    components = ma.fields.List(
        ma.fields.String(), metadata={"doc": "List of service components"}
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class ServiceModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, AppModelComponent]

    class ModelSchema(ma.Schema):
        service = ma.fields.Nested(
            ServiceClassSchema, metadata={"doc": "Service settings"}
        )
        service_config = ma.fields.Nested(
            ServiceConfigClassSchema,
            attribute="service-config",
            data_key="service-config",
            metadata={"doc": "Service config settings"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        profile_module = context["profile_module"]
        module = datatype.definition["module"]["qualified"]
        module_base_upper = datatype.definition["module"]["base-upper"]
        record_prefix = datatype.definition["module"]["prefix"]

        service_package = f"{module}.services.{profile_module}"

        config = set_default(datatype, "service-config", {})

        config.setdefault("generate", True)
        config.setdefault(
            "config-key",
            f"{module_base_upper}_{context['profile_upper']}_SERVICE_CONFIG",
        )
        config_module = config.setdefault(
            "module",
            f"{service_package}.config",
        )
        config.setdefault(
            "class",
            f"{config_module}.{record_prefix}ServiceConfig",
        )
        config.setdefault("extra-code", "")
        config.setdefault("service-id", datatype.definition["module"]["suffix-snake"])
        config.setdefault(
            "base-classes",
            ["PermissionsPresetsConfigMixin", "InvenioRecordServiceConfig"],
        )
        config.setdefault("components", [])
        append_array(
            datatype,
            "service-config",
            "imports",
            {
                "import": "invenio_records_resources.services.RecordServiceConfig",
                "alias": "InvenioRecordServiceConfig",
            },
        )
        append_array(
            datatype,
            "service-config",
            "imports",
            {"import": "oarepo_runtime.config.service.PermissionsPresetsConfigMixin"},
        )
        convert_config_to_qualified_name(config)

        service = set_default(datatype, "service", {})

        service.setdefault("generate", True)
        service.setdefault(
            "config-key",
            f"{module_base_upper}_{context['profile_upper']}_SERVICE_CLASS",
        )
        service.setdefault("proxy", "current_service")
        service_module = service.setdefault("module", f"{service_package}.service")
        service.setdefault("class", f"{service_module}.{record_prefix}Service")
        service.setdefault("extra-code", "")
        service.setdefault("base-classes", ["InvenioRecordService"])
        service.setdefault(
            "imports",
            [
                {
                    "import": "invenio_records_resources.services.RecordService",
                    "alias": "InvenioRecordService",
                }
            ],
        )
        convert_config_to_qualified_name(service)
